#import "TCFoundationExtensions.h"

@implementation TCDataScanner

+ (TCDataScanner *)scannerWithData:(NSData *)data
{
    return [[[self alloc] initWithData:data] autorelease];
}

- (id)initWithData:(NSData *)data
{
    if (self = [super init]) {
        _data = [data retain];
        _bytes = [data bytes];
        _offset = 0;
        _order = NSHostByteOrder();
        
        if ([data length] == 0) {
            [self dealloc];
            [NSException raise:NSRangeException format:@"attempt to data scan 0 length data"];
        }
    }

    return self;
}

- (void)dealloc
{
    [_data release];

    [super dealloc];
}

- (void)_checkLengthValidityAndRaise:(unsigned)length
{
    if ((_offset+length) > [_data length])
        [NSException raise:NSRangeException
                    format:@"requested offset extends beyond data's length"];
}

- (unsigned)offset
{
    return _offset;
}

- (void)setOffset:(unsigned)offset
{
    _offset = offset;
}

- (void)advanceOffsetBy:(unsigned)delta
{
    [self _checkLengthValidityAndRaise:delta];
    _offset += delta;
}

- (unsigned)byteOrder
{
    return _order;
}

- (void)setByteOrder:(unsigned)order
{
    _order = order;
}

- (unsigned)remainingBytes
{
    return [_data length] - _offset;
}

- (BOOL)matchString:(NSString *)string length:(unsigned)length encoding:(NSStringEncoding)encoding
{
    return [string isEqualToString:[self scanStringOfLength:length encoding:encoding]];
}

- (NSString *)scanStringOfLength:(unsigned)length encoding:(NSStringEncoding)encoding
{
    const unsigned char *bytesOffset = (_bytes+_offset);
    NSData *data;
    
    [self _checkLengthValidityAndRaise:length];
    data = [NSData dataWithBytes:bytesOffset length:length];
    _offset += length;
    
    return [NSString stringWithData:data encoding:encoding];
}

- (BOOL)matchCString:(const char *)cString length:(unsigned)length
{
    unsigned index = 0;
    const unsigned char *bytesOffset = (_bytes+_offset);

    [self _checkLengthValidityAndRaise:length];
    while (index < length) {
        if (bytesOffset[index] != cString[index])
            return NO;
        index++;
    }
    _offset += length;

    return YES;
}

- (const char *)scanCStringOfLength:(unsigned)length
{
    char *cString = [[NSMutableData dataWithLength:length] mutableBytes];
    const unsigned char *bytesOffset = (_bytes+_offset);
    unsigned index = 0;

    [self _checkLengthValidityAndRaise:length];
    while (index < length) {
        cString[index] = bytesOffset[index];
        index++;
    }
    _offset += length;

    return cString;
}

- (BOOL)matchByte:(unsigned char)byte
{
    return ([self scanByte] == byte);
}

- (unsigned char)scanByte
{
    const unsigned char *bytesOffset;
    
    [self _checkLengthValidityAndRaise:1];
    bytesOffset = (_bytes+_offset);
    _offset++;

    return *bytesOffset;
}

- (BOOL)matchShort:(unsigned short)shorty
{
    return ([self scanShort] == shorty);
}

- (unsigned short)scanShort
{
    const unsigned char *bytesOffset;
    unsigned short s;

    [self _checkLengthValidityAndRaise:2];
    bytesOffset = (_bytes+_offset);

    if (_order == NS_BigEndian)
        s = NSSwapBigShortToHost(*((unsigned short *)bytesOffset));
    else
        s = NSSwapLittleShortToHost(*((unsigned short *)bytesOffset));

    _offset += sizeof(short);

    return s;
}

- (BOOL)matchLong:(unsigned long)aLong
{
    return ([self scanLong] == aLong);
}

- (unsigned long)scanLong
{
    const unsigned char *bytesOffset;
    unsigned short s;

    [self _checkLengthValidityAndRaise:4];
    bytesOffset = (_bytes+_offset);

    if (_order == NS_BigEndian)
        s = NSSwapBigLongToHost(*((unsigned long *)bytesOffset));
    else
        s = NSSwapLittleLongToHost(*((unsigned long *)bytesOffset));

    _offset += sizeof(long);

    return s;
}

@end
