#import "TCNetworking.h"

#import <libc.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>

DEFINE_NSSTRING(TCSocketAcceptorCompletionNotification);
DEFINE_NSSTRING(TCSocketAcceptorUserInfoExceptionNameItemKey);
DEFINE_NSSTRING(TCSocketAcceptorUserInfoExceptionReasonItemKey);
DEFINE_NSSTRING(TCSocketAcceptorSocketArrayItemKey);

@interface TCSocketAcceptor(Private)
- (void)_prepareToAcceptConnections;
@end

@implementation TCSocketAcceptor

+ (TCSocketAcceptor *)acceptorWithPorts:(NSArray *)ports
{
    return [[[self alloc] initWithPorts:ports] autorelease];
}

- (id)initWithPorts:(NSArray *)ports
{
    if (self = [super init]) {
        if ([ports count] > 0) {
            _portsForListening = [ports copy];
            _listenQueueLength = 8;
            _descriptors = [NSMutableArray array];
            
            [self _prepareToAcceptConnections];
        }
        else {
            [self dealloc];	// no ports? then do nothing...
            return nil;		// not reached
        }
    }

    return self;
}

- (void)dealloc
{
    [[NSRunLoop currentRunLoop] removePort:_dataPort forMode:NSDefaultRunLoopMode];
    [_dataPort release];

    [_portsForListening release];
    [_descriptors release];
    
    [super dealloc];
}

- (void)_prepareToAcceptConnections
{
    NSEnumerator *portEnumerator = [_portsForListening objectEnumerator];
    NSNumber *number;
    int descriptor = 0, flag = 1;
    struct sockaddr_in addr;

    while (number = [portEnumerator nextObject]) {
        if ((descriptor = socket(AF_INET, SOCK_STREAM, 0)) < 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);
        
        if (setsockopt(descriptor, SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof (flag)) < 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        memset (&addr, 0, sizeof addr);
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = htonl (INADDR_ANY);
        addr.sin_port = htons([number unsignedIntValue]);

        if (bind(descriptor, (struct sockaddr *)&addr, sizeof(addr)) < 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        if (listen(descriptor, _listenQueueLength) < 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        [_descriptors addObject:[NSNumber numberWithUnsignedInt:descriptor]];
    }
}

- (NSArray *)acceptConnections
{
    NSMutableArray *sockets = [NSMutableArray array];
    NSEnumerator *descriptorEnumerator = [_descriptors objectEnumerator];
    NSNumber *number;
    int maxFD = 0, navail;
    fd_set readfds;

    FD_ZERO (&readfds);
    while (number = [descriptorEnumerator nextObject]) {
        unsigned descriptorToCheck = [number unsignedIntValue];

        FD_SET(descriptorToCheck, &readfds);
        if (descriptorToCheck > maxFD)
            maxFD = descriptorToCheck;
    }

    while ((navail = select (maxFD + 1, &readfds, 0, 0, NULL)) > 0) {
        NSEnumerator *descriptorEnumerator = [_descriptors objectEnumerator];
        while (number = [descriptorEnumerator nextObject]) {
            unsigned descriptorToCheck = [number unsignedIntValue];

            if (FD_ISSET(descriptorToCheck, &readfds)) {
                int newfd;
                struct sockaddr_in saddr;
                int saddr_siz = sizeof(saddr), remotePort = 0;
                TCSocket *incomingSocket;
                NSHost *remoteAddress, *localAddress;

                if ((newfd = accept(descriptorToCheck, (struct sockaddr *)&saddr, &saddr_siz)) < 0)
                    TCRaiseExceptionWithUNIXErrorCode(errno);

                remoteAddress = [NSHost hostWithAddress:
                    [NSString stringWithCString:inet_ntoa(saddr.sin_addr)]];
                remotePort = saddr.sin_port;

                if (getsockname(newfd, (struct sockaddr *)&saddr, &saddr_siz) < 0)
                    TCRaiseExceptionWithUNIXErrorCode(errno);

                localAddress = [NSHost hostWithAddress:
                    [NSString stringWithCString:inet_ntoa(saddr.sin_addr)]];

                incomingSocket = [[TCSocket alloc] initWithLocalAddress:localAddress
                                                              localPort:saddr.sin_port
                                                          remoteAddress:remoteAddress
                                                             remotePort:remotePort
                                                             descriptor:newfd];

                [sockets addObject:[incomingSocket autorelease]];
            }
        }

        return sockets;
    }

    TCRaiseExceptionWithUNIXErrorCode(errno);
    return nil; // not reached
}

- (void)acceptConnectionsInBackgroundAndNotify
{
    // create our dataPort, if it's not yet created
    if (_dataPort == nil) {
        _dataPort = [[NSPort alloc] init];
        [_dataPort setDelegate:self];

        [[NSRunLoop currentRunLoop] addPort:_dataPort forMode:NSDefaultRunLoopMode];
    }

    [NSThread detachNewThreadSelector:@selector(threadSelector:)
                             toTarget:self
                           withObject:nil];
}

- (void)threadSelector:(NSArray *)arguments
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    TCOperation *operation = nil;
    
    NS_DURING {
        NSArray *sockets = [self acceptConnections];
        NSEnumerator *socketEnumerator = [sockets objectEnumerator];
        TCSocket *socket;

        while (socket = [socketEnumerator nextObject])
            [socket setCloseOnDealloc:NO];

        operation = [TCOperation operationWithObject:sockets target:self selector:@selector(a:)
                                            userInfo:[NSDictionary dictionary]];
    }
    NS_HANDLER {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
            [localException name], TCSocketAcceptorUserInfoExceptionNameItemKey,
            [localException reason], TCSocketAcceptorUserInfoExceptionReasonItemKey,
            nil];

        // Send a nil object
        operation = [TCOperation operationWithObject:nil target:self selector:@selector(a:)
                                            userInfo:userInfo];
    }
    NS_ENDHANDLER;

    if (![_dataPort sendObject:operation])
        NSLog(@"-[TCSocketAcceptor threadSelector]: couldn't send mach message for %@",
              [operation description]);

    [pool release];
    [NSThread exit];
}

- (void)handlePortMessage:(NSPortMessage *)message
{
    NSData *serializedObjectData = [[message components] objectAtIndex:0];
    TCOperation *operation = [NSUnarchiver unarchiveObjectWithData:serializedObjectData];
    NSMutableDictionary *dict = [[[operation userInfo] mutableCopy] autorelease];

    [dict setObject:[operation object] forKey:TCSocketAcceptorSocketArrayItemKey];

    // Post a global notification
    [[NSNotificationCenter defaultCenter]
        postNotificationName:TCSocketAcceptorCompletionNotification
                      object:self
                    userInfo:dict];
}

- (unsigned)queueLength
{
    return _listenQueueLength;
}

- (void)setQueueLength:(unsigned)length
{
    _listenQueueLength = length;
}

@end
