/*
 * TCSocketConnector.m
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

/* NeXT headers */
#import <Foundation/Foundation.h>

/* Project headers */
#import "TCNetworking.h"

/* UNIX-level headers */
#import <libc.h>

DEFINE_NSSTRING(TCSocketConnectorInvalidDestinationException);
DEFINE_NSSTRING(TCSocketConnectorCompletionNotification);
DEFINE_NSSTRING(TCSocketConnectorThreadDictionaryItemKey);
DEFINE_NSSTRING(TCSocketConnectorUserInfoExceptionNameItemKey);
DEFINE_NSSTRING(TCSocketConnectorUserInfoExceptionReasonItemKey);

@interface TCSocketConnector(Private)
- (void)threadSelector:(NSArray *)arguments;
@end

@implementation TCSocketConnector

+ (TCSocketConnector *)defaultConnector
{
    NSMutableDictionary *threadDictionary = [[NSThread currentThread] threadDictionary];

    if ([threadDictionary objectForKey:TCSocketConnectorThreadDictionaryItemKey] == nil)
        [threadDictionary setObject:[[self alloc] init]
                             forKey:TCSocketConnectorThreadDictionaryItemKey];

    return [threadDictionary objectForKey:TCSocketConnectorThreadDictionaryItemKey];
}

- (id)init
{
    if (self = [super init]) {
        _port = [[NSPort alloc] init];
        [_port setDelegate:self];

        _operationUserInfo = [[NSMutableDictionary alloc] init];

        [[NSRunLoop currentRunLoop] addPort:_port forMode:NSDefaultRunLoopMode];
    }

    return self;
}

- (void)dealloc
{
    [[NSRunLoop currentRunLoop] removePort:_port forMode:NSDefaultRunLoopMode];
    [_port release];
    [_operationUserInfo release];

    [super dealloc];
}

- (int)newDescriptor
{
    int fd;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        TCRaiseExceptionWithUNIXErrorCode(errno);
        return -1; /* not reached */
    }
    else
        return fd;
}

- (TCSocket *)socketConnectedToHost:(NSHost *)address port:(unsigned)port
{
    int	descriptor = [self newDescriptor];
    unsigned structsize = sizeof(struct sockaddr);
    struct sockaddr_in *serverAddress;
    struct sockaddr_in *clientAddress;
    NSHost *localAddress;

    /* check for invalid parameters */
    if (address == nil)
        [NSException raise:TCSocketConnectorInvalidDestinationException
                    format:@"nil host object passed to %@", NSStringFromSelector(_cmd)];
 
    serverAddress = (struct sockaddr_in *)[[NSMutableData dataWithLength:
        sizeof(struct sockaddr_in)] mutableBytes];
    clientAddress = (struct sockaddr_in *)[[NSMutableData dataWithLength:
        sizeof(struct sockaddr_in)] mutableBytes];

    serverAddress->sin_family = AF_INET;
    serverAddress->sin_addr.s_addr = inet_addr([[address address] cString]);
    serverAddress->sin_port = htons(port);

    clientAddress->sin_family = AF_INET;
    clientAddress->sin_addr.s_addr = htonl(INADDR_ANY);
    clientAddress->sin_port = htons(0);

    if (bind(descriptor, (struct sockaddr *)clientAddress, sizeof(struct sockaddr)) < 0)
        TCRaiseExceptionWithUNIXErrorCode(errno);

    if (connect(descriptor, (struct sockaddr *)serverAddress, sizeof(struct sockaddr)) < 0)
        TCRaiseExceptionWithUNIXErrorCode(errno);

    if (getsockname(descriptor, (struct sockaddr *)clientAddress, &structsize) < 0)
        localAddress = [NSHost currentHost];
    else
        localAddress = [NSHost hostWithAddress:
            [NSString stringWithCString:inet_ntoa(clientAddress->sin_addr)]];

    return [[[TCSocket alloc] initWithLocalAddress:localAddress
                                         localPort:ntohs(clientAddress->sin_port)
                                     remoteAddress:address
                                        remotePort:port
                                        descriptor:descriptor] autorelease];
}

- (TCSocket *)socketConnectedToHostName:(NSString *)name port:(unsigned)port
{
    return [self socketConnectedToHost:[NSHost hostWithName:name] port:port];
}

- (void)detachNewConnectorAddress:(NSHost *)address port:(unsigned)port object:(id)anObject selector:(SEL)aSelector
{
    NSArray *arguments;

    arguments = [NSArray arrayWithObjects:address,
        [NSNumber numberWithUnsignedInt:port],
        [TCOperation operationWithObject:nil target:anObject selector:aSelector],
        nil];

    [NSThread detachNewThreadSelector:@selector(threadSelector:)
                             toTarget:self
                           withObject:[arguments retain]];
}

- (NSDictionary *)operationUserInfo
{
    return _operationUserInfo;
}

@end

@implementation TCSocketConnector(Private)

- (void)threadSelector:(NSArray *)arguments
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSHost *address = [arguments objectAtIndex:0];
    unsigned port = [[arguments objectAtIndex:1] unsignedIntValue];
    TCOperation *operation = [arguments objectAtIndex:2];
    TCSocket *socket;
    TCOperation *secondOperation;

    NS_DURING {
        socket = [self socketConnectedToHost:address port:port];
        [socket setCloseOnDealloc:NO];
        
        secondOperation = [TCOperation operationWithObject:socket
                                                    target:[operation target]
                                                  selector:[operation selector]];
    }
    NS_HANDLER {
        [_operationUserInfo setObject:[localException name]
                               forKey:TCSocketConnectorUserInfoExceptionNameItemKey];
        [_operationUserInfo setObject:[localException reason]
                               forKey:TCSocketConnectorUserInfoExceptionReasonItemKey];
        
        // Send a nil object
        secondOperation = [TCOperation operationWithObject:nil
                                                    target:[operation target]
                                                  selector:[operation selector]];
    }
    NS_ENDHANDLER;

    if (![_port sendObject:secondOperation])
        NSLog(@"-[TCSocketConnector threadSelector]: couldn't send mach message for %@",
              [secondOperation description]);

    [pool release];
    [NSThread exit];
}

- (void)handlePortMessage:(NSPortMessage *)message
{
    NSData *serializedObjectData = [[message components] objectAtIndex:0];
    TCOperation *operation = [NSUnarchiver unarchiveObjectWithData:serializedObjectData];

    // Executed the decoded operation
    [[operation target] performSelector:[operation selector] withObject:[operation object]];

    // Post a global notification
    [[NSNotificationCenter defaultCenter]
        postNotificationName:TCSocketConnectorCompletionNotification
                      object:[operation object]
                    userInfo:_operationUserInfo];

    [_operationUserInfo release];
    _operationUserInfo = nil;
}

@end
