/*
 * TCStream.m
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

/* NeXT headers */
#import <Foundation/Foundation.h>

/* Project headers */
#import "TCNetworking.h"

/* UNIX-level headers */
#import <libc.h>
#import <objc/objc-class.h>

#define _C_LNGLNG    'q'
#define _C_ULNGLNG    'Q'

@implementation TCStream

+ (TCStream *)streamWithFileDescriptor:(int)descriptor
{
    return [[[self alloc] initWithFileDescriptor:descriptor] autorelease];
}

- (id)init
{
    return nil;
}

- (id)initWithFileDescriptor:(int)descriptor
{
    if (self = [super init]) {
        _descriptor = descriptor;
        _endianness = NSHostByteOrder();

        [self setCloseOnDealloc:YES];
        [self setDebug:NO];
    }

    return self;
}

- (void)dealloc
{
    if (_streamFlags.closeOnDealloc) {
        if (_streamFlags.debug)
            NSLog(@"%@ closing on -dealloc", [self description]);
        
        close(_descriptor);
    }
    
    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"TCStream descriptor=%d", _descriptor];
}

- (id)initWithCoder:(NSCoder *)decoder
{
    if (self = [super init]) {
        [decoder decodeValueOfObjCType:@encode(int) at:&_descriptor];
        [decoder decodeValueOfObjCType:@encode(unsigned) at:&_streamFlags];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeValueOfObjCType:@encode(int) at:&_descriptor];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_streamFlags];
}

- (BOOL)debug
{
    return _streamFlags.debug;
}

- (void)setDebug:(BOOL)aFlag
{
    _streamFlags.debug = aFlag;
}

- (BOOL)closeOnDealloc
{
    return _streamFlags.closeOnDealloc;
}

- (void)setCloseOnDealloc:(BOOL)aFlag
{
    _streamFlags.closeOnDealloc = aFlag;
}

- (unsigned)availableBytes
{
    unsigned availableBytes;

    if (ioctl(_descriptor, FIONREAD, &availableBytes)!= -1) {
        if (_streamFlags.debug)
            NSLog(@"-[%@ availableBytes] %u bytes available", [self description], availableBytes);
        
        return availableBytes;
    }

    TCRaiseExceptionWithUNIXErrorCode(errno);
    return 0; /* not reached */
}

- (int)fileDescriptor
{
    return _descriptor;
}

- (void)closeStream
{
    close(_descriptor);
}

- (NSData *)readDataOfLength:(unsigned)length
{
    NSMutableData *data = [NSMutableData dataWithLength:length];
    unsigned bytesRead = 0;

    while (bytesRead < length) {
        int bytesReadPerPass = read(_descriptor, [data mutableBytes]+bytesRead, length-bytesRead);

        if (bytesReadPerPass == -1 || bytesReadPerPass == 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        bytesRead += bytesReadPerPass;
    }

    return data;
}

- (void)readDataOfLength:(unsigned)length at:(void *)ptr
{
    unsigned bytesRead = 0;

    if (ptr == NULL) {
        // this is so inefficient
        char dummy;
        unsigned i;
        for (i = 0; i < length; ++i) {
            int retVal = read(_descriptor, &dummy, 1);
            if (retVal == -1 || retVal == 0)
                TCRaiseExceptionWithUNIXErrorCode(errno);
        }
        
        return;
    }

    while (bytesRead < length) {
        int bytesReadPerPass = read(_descriptor, ptr+bytesRead, length-bytesRead);

        if (bytesReadPerPass == -1 || bytesReadPerPass == 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        bytesRead += bytesReadPerPass;
    }
}

- (NSData *)readData
{
    NSMutableData *data = [NSMutableData dataWithLength:1];
    
    // block until data is available
    if (read(_descriptor, [data mutableBytes], 1) != 0) {
        [data setLength:([self availableBytes]+1)];
        read(_descriptor, [data mutableBytes]+1, [self availableBytes]);
        return data;
    }
    else
        return [NSData data];
}

- (unsigned)writeData:(NSData *)data
{
    unsigned bytesWritten = 0;

    while (bytesWritten < [data length]) {
        int bytesWrittenPerPass = write(_descriptor,
                                        [data bytes]+bytesWritten,
                                        [data length]-bytesWritten);

        if (bytesWrittenPerPass == -1 || bytesWrittenPerPass == 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);
        
        bytesWritten += bytesWrittenPerPass;
    }

    return bytesWritten;
}

- (void)writeDataOfLength:(unsigned)length at:(void *)ptr
{
    unsigned bytesWritten = 0;

    while (bytesWritten < length) {
        int bytesWrittenPerPass = write(_descriptor, ptr+bytesWritten, length-bytesWritten);

        if (bytesWrittenPerPass == -1 || bytesWrittenPerPass == 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        bytesWritten += bytesWrittenPerPass;
    }
}

@end

@implementation TCStream(TypedStream)

- (void)setEndianness:(unsigned)order
{
    _endianness = order;
}

- (unsigned)endianness
{
    return _endianness;
}

- (void)writeString:(NSString *)string
{
    [self writeDataOfLength:[string cStringLength] at:[string cString]];
}

- (void)writeFormat:(NSString *)format, ...
{
    va_list	argList;
    NSString   *string;

    va_start(argList, format);
    string = [[NSString alloc] initWithFormat:format arguments:argList];
    [self writeDataOfLength:[string cStringLength] at:[string cString]];
    va_end(argList);
}

- (NSString *)readString
{
    return [NSString stringWithData:[self readData] encoding:NSASCIIStringEncoding];
}

- (NSString *)readLine
{
    NSMutableData *data = [NSMutableData dataWithLength:1024];
    unsigned counter = 0;
    char *dataPtr = [data mutableBytes];

    do {
        int status = read(_descriptor, dataPtr, 1);
        if (status == -1 || status == 0)
            TCRaiseExceptionWithUNIXErrorCode(errno);

        if (++counter == [data length])
            [data increaseLengthBy:1024];
    } while (*dataPtr++ != '\n');

    [data setLength:counter];
    return [NSString stringWithData:data encoding:NSASCIIStringEncoding];
}

- (int)readInt
{
    int i;
    
    [self readDataOfLength:sizeof(int) at:&i];
    i = (_endianness == NS_BigEndian) ? NSSwapBigIntToHost(i) : NSSwapLittleIntToHost(i);

    return i;
}

- (void)writeInt:(int)anInt
{
    int i = ((_endianness == NS_BigEndian) ? NSSwapHostIntToBig(anInt) :
             NSSwapHostIntToLittle(anInt));

    [self writeDataOfLength:sizeof(int) at:&i];
}

- (unsigned int)readUnsignedInt
{
    unsigned int i;

    [self readDataOfLength:sizeof(int) at:&i];
    i = (_endianness == NS_BigEndian) ? NSSwapBigIntToHost(i) : NSSwapLittleIntToHost(i);

    return i;
}

- (void)writeUnsignedInt:(unsigned int)anInt
{
    unsigned int i = (_endianness == NS_BigEndian) ? NSSwapHostIntToBig(anInt) :
             NSSwapHostIntToLittle(anInt);

    [self writeDataOfLength:sizeof(unsigned int) at:&i];
}

- (short)readShort
{
    short s;

    [self readDataOfLength:sizeof(short) at:&s];
    s = (_endianness == NS_BigEndian) ? NSSwapBigShortToHost(s) : NSSwapLittleShortToHost(s);

    return s;
}

- (void)writeShort:(short)aShort
{
    short s = (_endianness == NS_BigEndian) ? NSSwapHostShortToBig(aShort) :
               NSSwapHostShortToLittle(aShort);

    [self writeDataOfLength:sizeof(short) at:&s];
}

- (unsigned short)readUnsignedShort
{
    unsigned short s;

    [self readDataOfLength:sizeof(short) at:&s];
    s = (_endianness == NS_BigEndian) ? NSSwapBigShortToHost(s) : NSSwapLittleShortToHost(s);

    return s;
}

- (void)writeUnsignedShort:(unsigned short)aShort
{
    unsigned short s = ((_endianness == NS_BigEndian) ? NSSwapHostShortToBig(aShort) :
               NSSwapHostShortToLittle(aShort));

    [self writeDataOfLength:sizeof(unsigned short) at:&s];
}

- (unsigned char)readByte
{
    unsigned char c;

    [self readDataOfLength:sizeof(unsigned char) at:&c];
    return c;
}

- (void)writeByte:(unsigned char)aByte
{
    [self writeDataOfLength:sizeof(unsigned char) at:&aByte];
}

- (void)readValueOfObjCType:(const char *)type at:(void *)ptr
{
    switch(type[0]) {
        case _C_CHR:
        case _C_UCHR:
            [self readDataOfLength:sizeof(char) at:ptr];
            break;

        case _C_SHT: {
            short s;
            [self readDataOfLength:sizeof(short) at:&s];
            if (ptr != NULL)
                *((short *)ptr) = NSSwapBigShortToHost(s);
            break;
        }
        case _C_USHT:  {
            unsigned short s;
            [self readDataOfLength:sizeof(short) at:&s];
            if (ptr != NULL)
                *((short *)ptr) = NSSwapBigShortToHost(s);
            break;
        }

        case _C_INT:
        case _C_LNG: {
            int i;
            [self readDataOfLength:sizeof(int) at:&i];
            if (ptr != NULL)
                *((int *)ptr) = NSSwapBigIntToHost(i);
            break;
        }
        case _C_UINT:
        case _C_ULNG: {
            unsigned int i;
            [self readDataOfLength:sizeof(int) at:&i];
            if (ptr != NULL)
                *((int *)ptr) = NSSwapBigIntToHost(i);
            break;
        }

        case _C_FLT: {
            union floatFixer {
                unsigned char bytes[4];
                float f;
            } fixer;

            [self readDataOfLength:sizeof(float) at:&fixer.f];

            if (ptr != NULL)
                break;
            else if (NSHostByteOrder() != NS_BigEndian) {
                union floatFixer f2;

                f2.bytes[0] = fixer.bytes[3];
                f2.bytes[1] = fixer.bytes[2];
                f2.bytes[2] = fixer.bytes[1];
                f2.bytes[3] = fixer.bytes[0];

                *((float *)ptr) = f2.f;
            }
            else
                *((float *)ptr) = fixer.f;

            break;
        }

        case _C_LNGLNG:{
            long long l;
            [self readDataOfLength:sizeof(long long) at:&l];
            if (ptr != NULL)
                *((long long *)ptr) = NSSwapBigLongLongToHost(l);
            break;
        }
        case _C_ULNGLNG: {
            unsigned long long l;
            [self readDataOfLength:sizeof(long long) at:&l];
            if (ptr != NULL)
                *((unsigned long long *)ptr) = NSSwapBigLongLongToHost(l);
            break;
        }

        case _C_DBL:{
            union doubleFixer {
                unsigned char bytes[8];
                double d;
            } fixer;

            [self readDataOfLength:sizeof(double) at:&fixer.d];

            if (ptr != NULL)
                break;
            else if (NSHostByteOrder() != NS_BigEndian) {
                union doubleFixer d2;

                d2.bytes[0] = fixer.bytes[7];
                d2.bytes[1] = fixer.bytes[6];
                d2.bytes[2] = fixer.bytes[5];
                d2.bytes[3] = fixer.bytes[4];
                d2.bytes[4] = fixer.bytes[3];
                d2.bytes[5] = fixer.bytes[2];
                d2.bytes[6] = fixer.bytes[1];
                d2.bytes[7] = fixer.bytes[0];

                *((double *)ptr) = d2.d;
            }
            else
                *((double *)ptr) = fixer.d;

            break;
        }
            
        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"attempt to decode unknown or unsupported filer type %s", type];
    }
}

- (void)writeValueOfObjCType:(const char *)type at:(const void *)ptr
{
    if (ptr == NULL)
        return;
    
    switch(type[0]) {
        case _C_CHR:
        case _C_UCHR:
            [self writeDataOfLength:sizeof(char) at:(void *)ptr];
            break;

        case _C_SHT: {
            short s = NSSwapHostShortToBig(*((short *)ptr));
            [self writeDataOfLength:sizeof(short) at:&s];
            break;
        }
        case _C_USHT: {
            unsigned short s = NSSwapHostShortToBig(*((unsigned short *)ptr));
            [self writeDataOfLength:sizeof(short) at:&s];
            break;
        }

        case _C_INT:
        case _C_LNG: {
            int i = NSSwapHostIntToBig(*((int *)ptr));
            [self writeDataOfLength:sizeof(int) at:&i];
            break;
            }
        case _C_UINT:
        case _C_ULNG: {
            unsigned int i = NSSwapHostIntToBig(*((unsigned int *)ptr));
            [self writeDataOfLength:sizeof(int) at:&i];
            break;
        }

        case _C_FLT: {
            union floatFixer {
                unsigned char bytes[4];
                float f;
            } fixer;

            fixer.f = *((float *)ptr);

            if (NSHostByteOrder() != NS_BigEndian) {
                union floatFixer f2;

                f2.bytes[0] = fixer.bytes[3];
                f2.bytes[1] = fixer.bytes[2];
                f2.bytes[2] = fixer.bytes[1];
                f2.bytes[3] = fixer.bytes[0];

                [self writeDataOfLength:sizeof(float) at:&f2.f];
            }
            else
                [self writeDataOfLength:sizeof(float) at:&fixer.f];

            break;
        }

        case _C_LNGLNG:{
            long long l = NSSwapHostLongLongToBig(*((long long *)ptr));
            [self writeDataOfLength:sizeof(long long) at:&l];
            break;
        }
        case _C_ULNGLNG: {
            unsigned long long l = NSSwapHostLongLongToBig(*((unsigned long long *)ptr));
            [self writeDataOfLength:sizeof(long long) at:&l];
            break;
        }

        case _C_DBL:{
            union doubleFixer {
                unsigned char bytes[8];
                double d;
            } fixer;

            fixer.d = *((double *)ptr);

            if (NSHostByteOrder() != NS_BigEndian) {
                union doubleFixer d2;

                d2.bytes[0] = fixer.bytes[7];
                d2.bytes[1] = fixer.bytes[6];
                d2.bytes[2] = fixer.bytes[5];
                d2.bytes[3] = fixer.bytes[4];
                d2.bytes[4] = fixer.bytes[3];
                d2.bytes[5] = fixer.bytes[2];
                d2.bytes[6] = fixer.bytes[1];
                d2.bytes[7] = fixer.bytes[0];

                [self writeDataOfLength:sizeof(double) at:&d2.d];
            }
            else
                [self writeDataOfLength:sizeof(double) at:&fixer.d];

            break;
        }

       default:
            [NSException raise:NSInvalidArgumentException
                        format:@"attempt to encode unknown or unsupported filer type %s", type];
    }
}

@end

